import click

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, pretty_json
from zmon_cli.output import Output


@cli.command()
@click.argument('alert_id')
@click.argument('entity_ids', nargs=-1)
@click.pass_obj
@yaml_output_option
@pretty_json
def data(obj, alert_id, entity_ids, output, pretty):
    """Get check data for alert and entities"""
    client = get_client(obj.config)

    with Output('Retrieving alert data ...', nl=True, output=output, pretty_json=pretty) as act:
        data = client.get_alert_data(alert_id)

        if not entity_ids:
            result = data
        else:
            result = [d for d in data if d['entity'] in entity_ids]

        values = {v['entity']: v['results'][0]['value'] for v in result if len(v['results'])}

        act.echo(values)



import yaml

import click

from clickclick import AliasedGroup, Action, ok

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, pretty_json
from zmon_cli.output import Output
from zmon_cli.client import ZmonArgumentError


@cli.group('grafana', cls=AliasedGroup)
@click.pass_obj
def grafana(obj):
    """Manage Grafana dashboards"""
    pass


@grafana.command('get')
@click.argument('dashboard_id', type=click.STRING)
@click.pass_obj
@yaml_output_option
@pretty_json
def grafana_get(obj, dashboard_id, output, pretty):
    """Get ZMON grafana dashboard"""
    client = get_client(obj.config)

    with Output('Retrieving grafana dashboard ...', nl=True, output=output, pretty_json=pretty) as act:
        dashboard = client.get_grafana_dashboard(dashboard_id)
        act.echo(dashboard)


@grafana.command('update')
@click.argument('yaml_file', type=click.File('rb'))
@click.pass_obj
def grafana_update(obj, yaml_file):
    """Create/Update a single ZMON dashboard"""
    dashboard = yaml.safe_load(yaml_file)

    title = dashboard.get('dashboard', {}).get('title', '')

    client = get_client(obj.config)

    with Action('Updating dashboard {} ...'.format(title), nl=True) as act:
        try:
            client.update_grafana_dashboard(dashboard)
            ok(client.grafana_dashboard_url(dashboard))
        except ZmonArgumentError as e:
            act.error(e)


@grafana.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



import os
import logging

import yaml
import click
import clickclick
import zign.api
import requests

from clickclick import Action, error


DEFAULT_CONFIG_FILE = '~/.zmon-cli.yaml'
DEFAULT_TIMEOUT = 10


def configure_logging(loglevel):
    # configure file logger to not clutter stdout with log lines
    logging.basicConfig(level=loglevel, filename='/tmp/zmon-cli.log',
                        format='%(asctime)s %(levelname)s %(name)s: %(message)s')
    logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARNING)


def get_config_data(config_file=DEFAULT_CONFIG_FILE):
    fn = os.path.expanduser(config_file)
    data = {}

    try:
        if os.path.exists(fn):
            with open(fn) as fd:
                data = yaml.safe_load(fd)
        else:
            clickclick.warning('No configuration file found at [{}]'.format(config_file))

            data['url'] = click.prompt('ZMON Base URL (e.g. https://zmon.example.org/api/v1)')

            # TODO: either ask for fixed token or Zign
            data['user'] = click.prompt('ZMON username', default=os.environ['USER'])

            with open(fn, mode='w') as fd:
                yaml.dump(data, fd, default_flow_style=False,
                          allow_unicode=True,
                          encoding='utf-8')
    except Exception as e:
        error(e)

    return validate_config(data)


def set_config_file(config_file, default_url):
    while True:
        url = click.prompt('Please enter the ZMON base URL (e.g. https://demo.zmon.io)', default=default_url)

        with Action('Checking {}..'.format(url)):
            requests.get(url, timeout=5, allow_redirects=False)
            break

    data = {'url': url}

    fn = os.path.expanduser(config_file)
    with Action('Writing configuration to {}..'.format(fn)):
        with open(fn, 'w') as fd:
            yaml.safe_dump(data, fd, default_flow_style=False)


def validate_config(data):
    """
    >>> validate_config({'url': 'foo', 'token': '123'})['url']
    'foo'
    """
    if not data.get('url'):
        raise Exception('Config file improperly configured: key "url" is missing')

    if 'token' not in data:
        data['token'] = zign.api.get_token('zmon', ['uid'])

    return data



import yaml

import click

from clickclick import AliasedGroup, Action, ok

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, pretty_json
from zmon_cli.output import dump_yaml, Output


@cli.group('dashboard', cls=AliasedGroup)
@click.pass_obj
def dashboard(obj):
    """Manage ZMON dashboards"""
    pass


@dashboard.command('init')
@click.argument('yaml_file', type=click.File('wb'))
@click.pass_obj
def init(obj, yaml_file):
    """Initialize a new dashboard YAML file"""
    name = click.prompt('Dashboard name', default='Example dashboard')
    alert_teams = click.prompt('Alert Teams (comma separated)', default='Team1, Team2')

    user = obj.config.get('user', 'unknown')

    data = {
        'id': '',
        'name': name,
        'last_modified_by': user,
        'alert_teams': [t.strip() for t in alert_teams.split(',')],
        'tags': [],
        'view_mode': 'FULL',
        'shared_teams': [],
        'widget_configuration': [],
    }

    yaml_file.write(dump_yaml(data).encode('utf-8'))
    ok()


@dashboard.command('get')
@click.argument("dashboard_id", type=int)
@click.pass_obj
@yaml_output_option
@pretty_json
def dashboard_get(obj, dashboard_id, output, pretty):
    """Get ZMON dashboard"""
    client = get_client(obj.config)
    with Output('Retrieving dashboard ...', nl=True, output=output, pretty_json=pretty) as act:
        dashboard = client.get_dashboard(dashboard_id)
        act.echo(dashboard)


@dashboard.command('update')
@click.argument('yaml_file', type=click.Path(exists=True))
@click.pass_obj
def dashboard_update(obj, yaml_file):
    """Create/Update a single ZMON dashboard"""
    client = get_client(obj.config)
    dashboard = {}
    with open(yaml_file, 'rb') as f:
        dashboard = yaml.safe_load(f)

    msg = 'Creating new dashboard ...'
    if 'id' in dashboard:
        msg = 'Updating dashboard {} ...'.format(dashboard.get('id'))

    with Action(msg, nl=True):
        dash_id = client.update_dashboard(dashboard)
        ok(client.dashboard_url(dash_id))


@dashboard.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



import time

import click

from clickclick import AliasedGroup

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, pretty_json
from zmon_cli.output import Output
from zmon_cli.client import ZmonArgumentError


@cli.group('downtimes', cls=AliasedGroup)
def downtimes():
    """Manage downtimes"""
    pass


@downtimes.command('create')
@click.argument('entity_ids', nargs=-1)
@click.option('-d', '--duration', type=int, help='downtime duration in minutes', default=60)
@click.option('-c', '--comment')
@click.pass_obj
@yaml_output_option
@pretty_json
def create_downtime(obj, entity_ids, duration, comment, output, pretty):
    """Create downtime for specified entities"""
    client = get_client(obj.config)

    start_ts = time.time()
    end_ts = time.time() + (duration * 60)

    downtime = {
        'entities': entity_ids,
        'comment': comment or 'downtime by ZMON CLI',
        'start_time': start_ts,
        'end_time': end_ts
    }

    with Output('Creating downtime ...', nl=True, output=output, pretty_json=pretty) as act:
        try:
            new_downtime = client.create_downtime(downtime)
            act.echo(new_downtime)
        except ZmonArgumentError as e:
            act.error(str(e))


@downtimes.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



from datetime import datetime

import click

from clickclick import AliasedGroup, Action, ok

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, pretty_json
from zmon_cli.output import Output


########################################################################################################################
# TOKENS
########################################################################################################################

@cli.group('onetime-tokens', cls=AliasedGroup)
@click.pass_obj
def tv_tokens(obj):
    """Manage onetime tokens for Monitors/View only login"""
    pass


@tv_tokens.command('get')
@click.pass_obj
def get_tv_token(obj):
    """Retrieve a new token"""
    client = get_client(obj.config)

    with Action('Retrieving new one-time token ...', nl=True):
        token = client.get_onetime_token()
        ok(client.token_login_url(token.strip('"')))


@tv_tokens.command('list')
@click.pass_obj
@yaml_output_option
@pretty_json
def list_tv_token(obj, output, pretty):
    """List onetime tokens for your user"""
    client = get_client(obj.config)

    with Output('Retrieving onetime tokens ...', nl=True, output=output, pretty_json=pretty) as act:
        tokens = client.list_onetime_tokens()

        for t in tokens:
            t['created'] = datetime.fromtimestamp(t['created'] / 1000)
            if t['bound_at'] is not None:
                t['bound_at'] = datetime.fromtimestamp(t['bound_at'] / 1000)

        act.echo(tokens)



import click

from zmon_cli.cmds.command import cli, get_client, output_option, pretty_json
from zmon_cli.output import Output, render_search

from zmon_cli.client import ZmonArgumentError


@cli.command()
@click.argument('search_query', default="")
@click.option('--team', '-t', multiple=True, required=False,
              help='Filter search by team. Multiple teams filtering is supported.')
@click.option('--limit', '-l', multiple=False, required=False,
              help='Limit number of results, default is 25')
@click.pass_obj
@output_option
@pretty_json
def search(obj, search_query, team, limit, output, pretty):
    """
    Search dashboards, alerts, checks and grafana dashboards.

    Example:

        $ zmon search "search query" -t team-1 -t team-2
    """
    client = get_client(obj.config)

    with Output('Searching ...', nl=True, output=output, pretty_json=pretty, printer=render_search) as act:
        try:
            data = client.search(search_query, limit=limit, teams=team)

            for check in data['checks']:
                check['link'] = client.check_definition_url(check)

            for alert in data['alerts']:
                alert['link'] = client.alert_details_url(alert)

            for dashboard in data['dashboards']:
                dashboard['link'] = client.dashboard_url(dashboard['id'])

            for dashboard in data['grafana_dashboards']:
                dashboard['link'] = client.grafana_dashboard_url(dashboard)

            act.echo(data)
        except ZmonArgumentError as e:
            act.error(str(e))



import ast
import logging
import json
import functools
import re
import traceback

import requests
from requests.adapters import HTTPAdapter

from datetime import datetime
from urllib.parse import urljoin, urlsplit, urlunsplit, SplitResult

from opentracing_utils import trace, extract_span_from_kwargs
from urllib3 import Retry

from zmon_cli import __version__
from zmon_cli.config import DEFAULT_TIMEOUT


API_VERSION = 'v1'

ZMON_USER_AGENT = 'zmon-client/{}'.format(__version__)

ACTIVE_ALERT_DEF = 'checks/all-active-alert-definitions'
ACTIVE_CHECK_DEF = 'checks/all-active-check-definitions'
ALERT_DATA = 'status/alert'
ALERT_DEF = 'alert-definitions'
CHECK_DEF = 'check-definitions'
DASHBOARD = 'dashboard'
DOWNTIME = 'downtimes'
ENTITIES = 'entities'
GRAFANA = 'visualization/dashboards'
GROUPS = 'groups'
MEMBER = 'member'
PHONE = 'phone'
SEARCH = 'quick-search'
STATUS = 'status'
TOKENS = 'onetime-tokens'

ALERT_DETAILS_VIEW_URL = '#/alert-details/'
CHECK_DEF_VIEW_URL = '#/check-definitions/view/'
DASHBOARD_VIEW_URL = '#/dashboards/views/'
GRAFANA_DASHBOARD_URL = 'visualization/dashboard/'
TOKEN_LOGIN_URL = 'tv/'
BACKOFF_FACTOR = 0.3

logger = logging.getLogger(__name__)

parentheses_re = re.compile('[(]+|[)]+')
invalid_entity_id_re = re.compile('[^a-zA-Z0-9-@_.\\[\\]\\:]+')


class JSONDateEncoder(json.JSONEncoder):
    def default(self, obj):
        return obj.isoformat() if isinstance(obj, datetime) else super().default(obj)


class ZmonError(Exception):
    """ZMON client error."""

    def __init__(self, message=''):
        super().__init__('ZMON client error: {}'.format(message))


class ZmonArgumentError(ZmonError):
    """A ZMON client error indicating that a supplied object has missing or invalid attributes."""
    pass


def logged(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            logger.error('ZMON client failed in: {}'.format(f.__name__))
            raise

    return wrapper


def compare_entities(e1, e2):
    try:
        e1_copy = e1.copy()
        e1_copy.pop('last_modified', None)

        e2_copy = e2.copy()
        e2_copy.pop('last_modified', None)

        return (json.loads(json.dumps(e1_copy, cls=JSONDateEncoder)) ==
                json.loads(json.dumps(e2_copy, cls=JSONDateEncoder)))
    except Exception:
        # We failed during json serialiazation/deserialization, fallback to *not-equal*!
        logger.exception('Failed in `compare_entities`')
        return False


def get_valid_entity_id(e):
    return invalid_entity_id_re.sub('-', parentheses_re.sub(lambda m: '[' if '(' in m.group() else ']', e.lower()))


class Zmon:
    """ZMON client class that enables communication with ZMON backend.

    :param url: ZMON backend base url.
    :type url: str

    :param token: ZMON authentication token.
    :type token: str

    :param username: ZMON authentication username. Ignored if ``token`` is used.
    :type username: str

    :param password: ZMON authentication password. Ignored if ``token`` is used.
    :type password: str

    :param timeout: HTTP requests timeout. Default is 10 sec.
    :type timeout: int

    :param verify: Verify SSL connection. Default is ``True``.
    :type verify: bool

    :param user_agent: ZMON user agent. Default is generated by ZMON client and includes lib version.
    :type user_agent: str
    """

    def __init__(
        self,
        url,
        token=None,
        username=None,
        password=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_agent=ZMON_USER_AGENT,
        retry_count=None,
        *args,
        **kwargs,
    ):
        """Initialize ZMON client."""
        self.timeout = timeout

        split = urlsplit(url)
        self.base_url = urlunsplit(SplitResult(split.scheme, split.netloc, '', '', ''))
        self.url = urljoin(self.base_url, self._join_path(['api', API_VERSION, '']))

        self._session = requests.Session()
        if retry_count:
            retry = Retry(
                total=retry_count,
                read=retry_count,
                connect=retry_count,
                backoff_factor=BACKOFF_FACTOR,
            )
            adapter = HTTPAdapter(max_retries=retry)
            self._session.mount('http://', adapter)
            self._session.mount('https://', adapter)
        self._timeout = timeout
        self.user_agent = user_agent

        if username and password and token is None:
            self._session.auth = (username, password)

        self._session.headers.update({'User-Agent': user_agent, 'Content-Type': 'application/json'})

        if token:
            self._session.headers.update({'Authorization': 'Bearer {}'.format(token)})

        if not verify:
            logger.warning('ZMON client will skip SSL verification!')
            requests.packages.urllib3.disable_warnings()
            self._session.verify = False

    @property
    def session(self):
        return self._session

    @staticmethod
    def is_valid_entity_id(entity_id):
        return invalid_entity_id_re.search(entity_id) is None

    @staticmethod
    def validate_check_command(src):
        """
        Validates if ``check command`` is valid syntax. Raises exception in case of invalid syntax.

        :param src: Check command python source code.
        :type src: str

        :raises: ZmonError
        """
        try:
            ast.parse(src)
        except Exception as e:
            raise ZmonError('Invalid check command: {}'.format(e))

    def _join_path(self, parts):
        return '/'.join(str(p).strip('/') for p in parts)

    def endpoint(self, *args, trailing_slash=True, base_url=None):
        parts = list(args)

        # Ensure trailing slash!
        if trailing_slash:
            parts.append('')

        url = self.url if not base_url else base_url

        return urljoin(url, self._join_path(parts))

    def json(self, resp):
        resp.raise_for_status()

        return resp.json()

########################################################################################################################
# DEEPLINKS
########################################################################################################################

    def check_definition_url(self, check_definition: dict) -> str:
        """
        Return direct deeplink to check definition view on ZMON UI.

        :param check_definition: check_difinition dict.
        :type check_definition: dict

        :return: Deeplink to check definition view.
        :rtype: str
        """
        return self.endpoint(CHECK_DEF_VIEW_URL, check_definition['id'], base_url=self.base_url)

    def alert_details_url(self, alert: dict) -> str:
        """
        Return direct deeplink to alert details view on ZMON UI.

        :param alert: alert dict.
        :type alert: dict

        :return: Deeplink to alert details view.
        :rtype: str
        """
        return self.endpoint(ALERT_DETAILS_VIEW_URL, alert['id'], base_url=self.base_url)

    def dashboard_url(self, dashboard_id: int) -> str:
        """
        Return direct deeplink to ZMON dashboard.

        :param dashboard_id: ZMON Dashboard ID.
        :type dashboard_id: int

        :return: Deeplink to dashboard.
        :rtype: str
        """
        return self.endpoint(DASHBOARD_VIEW_URL, dashboard_id, base_url=self.base_url)

    def token_login_url(self, token: str) -> str:
        """
        Return direct deeplink to ZMON one-time login.

        :param token: One-time token.
        :type token: str

        :return: Deeplink to ZMON one-time login.
        :rtype: str
        """
        return self.endpoint(TOKEN_LOGIN_URL, token, base_url=self.base_url)

    def grafana_dashboard_url(self, dashboard: dict) -> str:
        """
        Return direct deeplink to Grafana dashboard.

        :param dashboard: Grafana dashboard dict.
        :type dashboard: dict

        :return: Deeplink to Grafana dashboard.
        :rtype: str
        """
        if dashboard.get('id', None):
            return self.endpoint(GRAFANA_DASHBOARD_URL, dashboard['id'], base_url=self.base_url, trailing_slash=False)
        return ""

    @logged
    def status(self) -> dict:
        """
        Return ZMON status from status API.

        :return: ZMON status.
        :rtype: dict
        """
        resp = self.session.get(self.endpoint(STATUS), timeout=self._timeout)

        return self.json(resp)

########################################################################################################################
# ENTITIES
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def get_entities(self, query=None, **kwargs) -> list:
        """
        Get ZMON entities, with optional filtering.

        :param query: Entity filtering query. Default is ``None``. Example query ``{'type': 'instance'}`` to return
                      all entities of type: ``instance``.
        :type query: dict

        :return: List of entities.
        :rtype: list
        """
        query_str = json.dumps(query) if query else ''
        logger.debug('Retrieving entities with query: {} ...'.format(query_str))

        current_span = extract_span_from_kwargs(**kwargs)
        current_span.log_kv({'query': query_str})

        params = {'query': query_str} if query else None

        resp = self.session.get(self.endpoint(ENTITIES), params=params, timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def get_entity(self, entity_id: str, **kwargs) -> str:
        """
        Retrieve single entity.

        :param entity_id: Entity ID.
        :type entity_id: str

        :return: Entity dict.
        :rtype: dict
        """
        logger.debug('Retrieving entities with id: {} ...'.format(entity_id))

        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('entity_id', entity_id)

        resp = self.session.get(self.endpoint(ENTITIES, entity_id, trailing_slash=False), timeout=self._timeout)
        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def add_entity(self, entity: dict, **kwargs) -> requests.Response:
        """
        Create or update an entity on ZMON.

        .. note::

            ZMON PUT entity API doesn't return JSON response.

        :param entity: Entity dict.
        :type entity: dict

        :return: Response object.
        :rtype: :class:`requests.Response`
        """
        if 'id' not in entity or 'type' not in entity:
            raise ZmonArgumentError('Entity "id" and "type" are required.')

        if not self.is_valid_entity_id(entity['id']):
            raise ZmonArgumentError('Invalid entity ID.')

        logger.debug('Adding new entity: {} ...'.format(entity['id']))

        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('entity_id', entity['id'])

        data = json.dumps(entity, cls=JSONDateEncoder)
        resp = self.session.put(self.endpoint(ENTITIES, trailing_slash=False), data=data, timeout=self._timeout)

        resp.raise_for_status()

        return resp

    @trace(pass_span=True)
    @logged
    def delete_entity(self, entity_id: str, **kwargs) -> bool:
        """
        Delete entity from ZMON.

        .. note::

            ZMON DELETE entity API doesn't return JSON response.

        :param entity_id: Entity ID.
        :type entity_id: str

        :return: True if succeeded, False otherwise.
        :rtype: bool
        """
        logger.debug('Removing existing entity: {} ...'.format(entity_id))

        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('entity_id', entity_id)

        resp = self.session.delete(self.endpoint(ENTITIES, entity_id))

        resp.raise_for_status()

        return resp.text == '1'

########################################################################################################################
# DASHBOARD
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def get_dashboard(self, dashboard_id: str, **kwargs) -> dict:
        """
        Retrieve a ZMON dashboard.

        :param dashboard_id: ZMON dashboard ID.
        :type dashboard_id: int, str

        :return: Dashboard dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('dashboard_id', dashboard_id)

        resp = self.session.get(self.endpoint(DASHBOARD, dashboard_id), timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def update_dashboard(self, dashboard: dict, **kwargs) -> dict:
        """
        Create or update dashboard.

        If dashboard has an ``id`` then dashboard will be updated, otherwise a new dashboard is created.

        :param dashboard: ZMON dashboard dict.
        :type dashboard: int, str

        :return: Dashboard dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if 'id' in dashboard and dashboard['id']:
            logger.debug('Updating dashboard with ID: {} ...'.format(dashboard['id']))
            current_span.set_tag('dashboard_id', dashboard['id'])

            resp = self.session.post(self.endpoint(DASHBOARD, dashboard['id']), json=dashboard, timeout=self._timeout)
        else:
            # new dashboard
            logger.debug('Adding new dashboard ...')
            resp = self.session.post(self.endpoint(DASHBOARD), json=dashboard, timeout=self._timeout)

        resp.raise_for_status()

        return self.json(resp)

########################################################################################################################
# CHECK-DEFS
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def get_check_definition(self, definition_id: int, **kwargs) -> dict:
        """
        Retrieve check defintion.

        :param defintion_id: Check defintion id.
        :type defintion_id: int

        :return: Check definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('check_id', definition_id)

        resp = self.session.get(self.endpoint(CHECK_DEF, definition_id), timeout=self._timeout)

        # TODO: total hack! API returns 200 if check def does not exist!
        if resp.text == '':
            resp.status_code = 404
            resp.reason = 'Not Found'

        return self.json(resp)

    @trace()
    @logged
    def get_check_definitions(self) -> list:
        """
        Return list of all ``active`` check definitions.

        :return: List of check-defs.
        :rtype: list
        """
        resp = self.session.get(self.endpoint(ACTIVE_CHECK_DEF), timeout=self._timeout)

        return self.json(resp).get('check_definitions')

    @trace(pass_span=True)
    @logged
    def update_check_definition(self, check_definition, skip_validation=False, **kwargs) -> dict:
        """
        Update existing check definition.

        Atrribute ``owning_team`` is required. If ``status`` is not set, then it will be set to ``ACTIVE``.

        :param check_definition: ZMON check definition dict.
        :type check_definition: dict

        :param skip_validation: Skip validation of the check command syntax.
        :type skip_validation: bool

        :return: Check definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if 'owning_team' not in check_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Check definition must have "owning_team"'})
            raise ZmonArgumentError('Check definition must have "owning_team"')

        if 'status' not in check_definition:
            check_definition['status'] = 'ACTIVE'

        if not skip_validation:
            try:
                self.validate_check_command(check_definition['command'])
            except Exception:
                current_span.set_tag('error', True)
                current_span.log_kv({'exception': traceback.format_exc()})
                raise

        resp = self.session.post(self.endpoint(CHECK_DEF), json=check_definition, timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def delete_check_definition(self, check_definition_id: int, **kwargs) -> requests.Response:
        """
        Delete existing check definition.

        :param check_definition_id: ZMON check definition ID.
        :type check_definition_id: int

        :return: HTTP response.
        :rtype: :class:`requests.Response`
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('check_id', str(check_definition_id))

        resp = self.session.delete(self.endpoint(CHECK_DEF, check_definition_id))

        resp.raise_for_status()

        return resp

########################################################################################################################
# ALERT-DEFS & DATA
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def get_alert_definition(self, alert_id: int, **kwargs) -> dict:
        """
        Retrieve alert definition.

        :param alert_id: Alert definition ID.
        :type alert_id: int

        :return: Alert definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('alert_id', str(alert_id))

        resp = self.session.get(self.endpoint(ALERT_DEF, alert_id), timeout=self._timeout)

        return self.json(resp)

    @trace()
    @logged
    def get_alert_definitions(self) -> list:
        """
        Return list of all ``active`` alert definitions.

        :return: List of alert-defs.
        :rtype: list
        """
        resp = self.session.get(self.endpoint(ACTIVE_ALERT_DEF), timeout=self._timeout)

        return self.json(resp).get('alert_definitions')

    @trace(pass_span=True)
    @logged
    def create_alert_definition(self, alert_definition: dict, **kwargs) -> dict:
        """
        Create new alert definition.

        Attributes ``last_modified_by`` and ``check_definition_id`` are required.
        If ``status`` is not set, then it will be set to ``ACTIVE``.

        :param alert_definition: ZMON alert definition dict.
        :type alert_definition: dict

        :return: Alert definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if 'last_modified_by' not in alert_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Alert definition must have "last_modified_by"'})
            raise ZmonArgumentError('Alert definition must have "last_modified_by"')

        if 'status' not in alert_definition:
            alert_definition['status'] = 'ACTIVE'

        if 'check_definition_id' not in alert_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Alert definition must have "last_modified_by"'})
            raise ZmonArgumentError('Alert defintion must have "check_definition_id"')
        current_span.set_tag('check_id', alert_definition['check_definition_id'])

        resp = self.session.post(self.endpoint(ALERT_DEF), json=alert_definition, timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def update_alert_definition(self, alert_definition: dict, **kwargs) -> dict:
        """
        Update existing alert definition.

        Atrributes ``id``, ``last_modified_by`` and ``check_definition_id`` are required.
        If ``status`` is not set, then it will be set to ``ACTIVE``.

        :param alert_definition: ZMON alert definition dict.
        :type alert_definition: dict

        :return: Alert definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if 'last_modified_by' not in alert_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Alert definition must have "last_modified_by"'})
            raise ZmonArgumentError('Alert definition must have "last_modified_by"')

        if 'id' not in alert_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Alert definition must have "id"'})
            raise ZmonArgumentError('Alert definition must have "id"')
        current_span.set_tag('alert_id', alert_definition['id'])

        if 'check_definition_id' not in alert_definition:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Alert definition must have "check_definition_id"'})
            raise ZmonArgumentError('Alert defintion must have "check_definition_id"')
        current_span.set_tag('check_id', alert_definition['check_definition_id'])

        if 'status' not in alert_definition:
            alert_definition['status'] = 'ACTIVE'

        resp = self.session.put(
            self.endpoint(ALERT_DEF, alert_definition['id']), json=alert_definition, timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def delete_alert_definition(self, alert_definition_id: int, **kwargs) -> dict:
        """
        Delete existing alert definition.

        :param alert_definition_id: ZMON alert definition ID.
        :type alert_definition_id: int

        :return: Alert definition dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('alert_id', str(alert_definition_id))
        resp = self.session.delete(self.endpoint(ALERT_DEF, alert_definition_id))

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def get_alert_data(self, alert_id: int, **kwargs) -> dict:
        """
        Retrieve alert data.

        Response is a ``dict`` with entity ID as a key, and check return value as a value.

        :param alert_id: ZMON alert ID.
        :type alert_id: int

        :return: Alert data dict.
        :rtype: dict

        Example:

        .. code-block:: json

            {
                "entity-id-1": 122,
                "entity-id-2": 0,
                "entity-id-3": 100
            }
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('alert_id', str(alert_id))
        resp = self.session.get(self.endpoint(ALERT_DATA, alert_id, 'all-entities'), timeout=self._timeout)

        return self.json(resp)

########################################################################################################################
# SEARCH
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def search(self, q, limit=None, teams=None, **kwargs) -> dict:
        """
        Search ZMON dashboards, checks, alerts and grafana dashboards with optional team filtering.

        :param q: search query.
        :type q: str

        :param teams: List of team IDs. Default is None.
        :type teams: list

        :return: Search result.
        :rtype: dict

        Example:

        .. code-block:: json

            {
                "alerts": [{"id": "123", "title": "ZMON alert", "team": "ZMON"}],
                "checks": [{"id": "123", "title": "ZMON check", "team": "ZMON"}],
                "dashboards": [{"id": "123", "title": "ZMON dashboard", "team": "ZMON"}],
                "grafana_dashboards": [{"id": "123", "title": "ZMON grafana", "team": ""}],
            }
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if teams and type(teams) not in (list, tuple):
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': '"teams" should be a list'})
            raise ZmonArgumentError('"teams" should be a list!')

        params = {'query': q}
        if limit:
            params.update({'limit': limit})
        if teams:
            params['teams'] = ','.join(teams)

        current_span.log_kv({'query', json.dumps(params)})
        resp = self.session.get(self.endpoint(SEARCH), params=params, timeout=self._timeout)

        return self.json(resp)


########################################################################################################################
# ONETIME-TOKENS
########################################################################################################################

    @trace()  # noqa: E301
    @logged
    def list_onetime_tokens(self) -> list:
        """
        List exisitng one-time tokens.

        :return: List of one-time tokens, with relevant attributes.
        :retype: list

        Example:

        .. code-block:: yaml

            - bound_at: 2016-09-08 14:00:12.645999
              bound_expires: 1503744673506
              bound_ip: 192.168.20.16
              created: 2016-08-26 12:51:13.506000
              token: 9pSzKpcO
        """
        resp = self.session.get(self.endpoint(TOKENS), timeout=self._timeout)

        return self.json(resp)

    @trace()
    @logged
    def get_onetime_token(self) -> str:
        """
        Retrieve new one-time token.

        You can use :func:`zmon_cli.client.Zmon.token_login_url` to return a deeplink to one-time login.

        :return: One-time token.
        :retype: str
        """
        resp = self.session.post(self.endpoint(TOKENS), json={}, timeout=self._timeout)

        resp.raise_for_status()

        return resp.text

########################################################################################################################
# GRAFANA
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def get_grafana_dashboard(self, grafana_dashboard_uid: str, **kwargs) -> dict:
        """
        Retrieve Grafana dashboard.

        :param grafana_dashboard_uid: Grafana dashboard UID.
        :type grafana_dashboard_uid: str

        :return: Grafana dashboard dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)
        current_span.set_tag('grafana_dashboard_uid', grafana_dashboard_uid)
        url = self.endpoint(GRAFANA, grafana_dashboard_uid, trailing_slash=False)
        resp = self.session.get(url, timeout=self._timeout)

        return self.json(resp)

    @trace(pass_span=True)
    @logged
    def update_grafana_dashboard(self, grafana_dashboard: dict, **kwargs) -> dict:
        """
        Update existing Grafana dashboard.

        Atrributes ``uid`` and ``title`` are required.

        :param grafana_dashboard: Grafana dashboard dict.
        :type grafana_dashboard: dict

        :return: Grafana dashboard dict.
        :rtype: dict
        """
        current_span = extract_span_from_kwargs(**kwargs)

        if 'uid' not in grafana_dashboard['dashboard']:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Grafana dashboard must have "uid". Use Grafana6 dashboard format.'})
            raise ZmonArgumentError('Grafana dashboard must have "uid". Hint: Use Grafana6 dashboard format.')

        elif 'title' not in grafana_dashboard['dashboard']:
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Grafana dashboard must have "title"'})
            raise ZmonArgumentError('Grafana dashboard must have "title"')

        current_span.set_tag('grafana_dashboard_uid', grafana_dashboard['dashboard']['uid'])

        if 'id' in grafana_dashboard['dashboard'] and grafana_dashboard['dashboard']['id'] is not None:
            current_span.set_tag('grafana_dashboard_id', grafana_dashboard['dashboard']['id'])

        req = {'dashboard': grafana_dashboard['dashboard'],
               'overwrite': True}

        if 'folderId' in grafana_dashboard.get('meta', {}):
            req['folderId'] = grafana_dashboard['meta']['folderId']

        resp = self.session.post(self.endpoint(GRAFANA), json=json.dumps(req), timeout=self._timeout)

        return self.json(resp)

########################################################################################################################
# DOWNTIMES
########################################################################################################################

    @trace(pass_span=True)
    @logged
    def create_downtime(self, downtime: dict, **kwargs) -> dict:
        """
        Create a downtime for specific entities.

        Atrributes ``entities`` list, ``start_time`` and ``end_time`` timestamps are required.

        :param downtime: Downtime dict.
        :type downtime: dict

        :return: Downtime dict.
        :rtype: dict

        Example downtime:

        .. code-block:: json

            {
                "entities": ["entity-id-1", "entity-id-2"],
                "comment": "Planned maintenance",
                "start_time": 1473337437.312921,
                "end_time": 1473341037.312921,
            }
        """
        current_span = extract_span_from_kwargs(**kwargs)
        if not downtime.get('entities'):
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'At least one entity ID should be specified'})
            raise ZmonArgumentError('At least one entity ID should be specified')

        if not downtime.get('start_time') or not downtime.get('end_time'):
            current_span.set_tag('error', True)
            current_span.log_kv({'exception': 'Downtime must specify "start_time" and "end_time"'})
            raise ZmonArgumentError('Downtime must specify "start_time" and "end_time"')

        current_span.set_tag('entity_ids', str(downtime.get('entities')))
        # FIXME - those also?
        # current_span.set_tag('start_time', str(downtime.get('start_time')))
        # current_span.set_tag('end_time', str(downtime.get('end_time')))

        resp = self.session.post(self.endpoint(DOWNTIME), json=downtime, timeout=self._timeout)

        return self.json(resp)

########################################################################################################################
# GROUPS - MEMBERS - ???
########################################################################################################################

    @logged
    def get_groups(self):
        resp = self.session.get(self.endpoint(GROUPS), timeout=self._timeout)

        return self.json(resp)

    @logged
    def switch_active_user(self, group_name, user_name):
        resp = self.session.delete(self.endpoint(GROUPS, group_name, 'active'))

        if not resp.ok:
            logger.error('Failed to de-activate group: {}'.format(group_name))
            resp.raise_for_status()

        logger.debug('Switching active user: {}'.format(user_name))

        resp = self.session.put(self.endpoint(GROUPS, group_name, 'active', user_name), timeout=self._timeout)

        if not resp.ok:
            logger.error('Failed to switch active user {}'.format(user_name))
            resp.raise_for_status()

        return resp.text == '1'

    @logged
    def add_member(self, group_name, user_name):
        resp = self.session.put(self.endpoint(GROUPS, group_name, MEMBER, user_name), timeout=self._timeout)

        resp.raise_for_status()

        return resp.text == '1'

    @logged
    def remove_member(self, group_name, user_name):
        resp = self.session.delete(self.endpoint(GROUPS, group_name, MEMBER, user_name))

        resp.raise_for_status()

        return resp.text == '1'

    @logged
    def add_phone(self, member_email, phone_nr):
        resp = self.session.put(self.endpoint(GROUPS, member_email, PHONE, phone_nr), timeout=self._timeout)

        resp.raise_for_status()

        return resp.text == '1'

    @logged
    def remove_phone(self, member_email, phone_nr):
        resp = self.session.delete(self.endpoint(GROUPS, member_email, PHONE, phone_nr))

        resp.raise_for_status()

        return resp.text == '1'

    @logged
    def set_name(self, member_email, member_name):
        resp = self.session.put(self.endpoint(GROUPS, member_email, PHONE, member_name), timeout=self._timeout)

        resp.raise_for_status()

        return resp



import click
import logging
import os

from clickclick import AliasedGroup
from easydict import EasyDict

from zmon_cli import __version__

from zmon_cli.config import DEFAULT_CONFIG_FILE, DEFAULT_TIMEOUT
from zmon_cli.config import get_config_data, configure_logging, set_config_file

from zmon_cli.output import Output, render_status

from zmon_cli.client import Zmon


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

output_option = click.option('-o', '--output', type=click.Choice(['text', 'json', 'yaml']), default='text',
                             help='Use alternative output format')

yaml_output_option = click.option('-o', '--output', type=click.Choice(['text', 'json', 'yaml']), default='yaml',
                                  help='Use alternative output format. Default is YAML.')

pretty_json = click.option('--pretty', is_flag=True,
                           help='Pretty print JSON output. Ignored if output format is not JSON')

retries = click.option(
    "--retries",
    default=None,
    help='Number of retries.'
)


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo('ZMON CLI {}'.format(__version__))
    ctx.exit()


def get_client(config, **kwargs):
    verify = config.get("verify", True)
    common_arguments = {
        "url": config["url"],
        "verify": verify,
        "timeout": config.get("timeout", DEFAULT_TIMEOUT),
        **kwargs,
    }

    if "user" in config and "password" in config:
        return Zmon(
            **common_arguments,
            username=config["user"],
            password=config["password"],
        )
    elif os.environ.get("ZMON_TOKEN"):
        return Zmon(
            **common_arguments,
            token=os.environ.get("ZMON_TOKEN"),
        )
    elif "token" in config:
        return Zmon(
            **common_arguments,
            token=config["token"],
        )

    raise RuntimeError("Failed to intitialize ZMON client. Invalid configuration!")


########################################################################################################################
# CLI
########################################################################################################################

@click.group(cls=AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.option('-c', '--config-file', help='Use alternative config file', default=DEFAULT_CONFIG_FILE, metavar='PATH')
@click.option('-v', '--verbose', help='Verbose logging', is_flag=True)
@click.option('-V', '--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True)
@click.option('-t', '--timeout', help='timeout for calls', default=DEFAULT_TIMEOUT)
@click.pass_context
def cli(ctx, config_file, verbose, timeout=DEFAULT_TIMEOUT):
    """
    ZMON command line interface
    """
    configure_logging(logging.DEBUG if verbose else logging.INFO)

    fn = os.path.expanduser(config_file)
    config = {}

    if os.path.exists(fn):
        config = get_config_data(config_file)

    config['timeout'] = timeout

    ctx.obj = EasyDict(config=config)


@cli.command()
@click.option('-c', '--config-file', help='Use alternative config file', default=DEFAULT_CONFIG_FILE, metavar='PATH')
@click.pass_obj
def configure(obj, config_file):
    """Configure ZMON URL and credentials"""
    set_config_file(config_file, obj.config.get('url'))


@cli.command()
@click.pass_obj
@output_option
@pretty_json
def status(obj, output, pretty):
    """Check ZMON system status"""
    client = get_client(obj.config)

    with Output('Retrieving status ...', printer=render_status, output=output, pretty_json=pretty) as act:
        status = client.status()
        act.echo(status)


@click.command()
@click.pass_context
def help(ctx):
    print(ctx.get_help())


cli.add_command(help)



import click

from clickclick import Action

from zmon_cli.cmds.command import cli, get_client


@cli.group(invoke_without_command=True)
@click.pass_context
def groups(ctx):
    """Manage contact groups"""
    client = get_client(ctx.obj.config)

    if not ctx.invoked_subcommand:
        with Action('Retrieving groups ...', nl=True) as act:
            groups = client.get_groups()

            if len(groups) == 0:
                act.warning('No groups found!')

            for g in groups:
                print('Name: {} Id: {}'.format(g['name'], g['id']))

                print('\tMembers:')
                for m in g['members']:
                    member = client.get_member(m)
                    print('\t\t{} {} {}'.format(member['name'], member['email'], member['phones']))

                print('\tActive:')
                for m in g['active']:
                    member = client.get_member(m)
                    print('\t\t{} {} {}'.format(member['name'], member['email'], member['phones']))


@groups.command('switch')
@click.argument('group_name')
@click.argument('user_name')
@click.pass_obj
def switch_active(obj, group_name, user_name):
    client = get_client(obj.config)

    with Action('Switching active user ...') as act:
        switched = client.switch_active_user(group_name, user_name)
        if not switched:
            act.error('Failed to switch')


@cli.group()
@click.pass_obj
def members(obj):
    """Manage group membership"""
    pass


@members.command('add')
@click.argument('group_name')
@click.argument('user_name')
@click.pass_obj
def member_add(obj, group_name, user_name):
    client = get_client(obj.config)

    with Action('Adding user ...') as act:
        added = client.add_member(group_name, user_name)

        if not added:
            act.error('Failed to add member')


@members.command('remove')
@click.argument('group_name')
@click.argument('user_name')
@click.pass_obj
def member_remove(obj, group_name, user_name):
    client = get_client(obj.config)

    with Action('Removing user ...') as act:
        removed = client.remove_member(group_name, user_name)

        if not removed:
            act.error('Failed to remove member')


@members.command('add-phone')
@click.argument('member_email')
@click.argument('phone_nr')
@click.pass_obj
def add_phone(obj, member_email, phone_nr):
    client = get_client(obj.config)

    with Action('Adding phone ...') as act:
        added = client.add_phone(member_email, phone_nr)

        if not added:
            act.error('Failed to add phone')


@members.command('remove-phone')
@click.argument('member_email')
@click.argument('phone_nr')
@click.pass_obj
def remove_phone(obj, member_email, phone_nr):
    client = get_client(obj.config)

    with Action('Removing phone number ...') as act:
        removed = client.remove_phone(member_email, phone_nr)

        if not removed:
            act.error('Failed to remove phone')


@members.command('change-name')
@click.argument('member_email')
@click.argument('member_name')
@click.pass_obj
def set_name(obj, member_email, member_name):
    client = get_client(obj.config)

    with Action('Changing user name ...'):
        client.set_name(member_email, member_name)


@members.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



import json
import time

import yaml
import calendar

from clickclick import print_table, OutputFormat, action, secho, error, ok, info


# fields to dump as literal blocks
LITERAL_FIELDS = set(['command', 'condition', 'description'])

# custom sorting of YAML fields (i.e. we are not using the default lexical YAML ordering)
FIELD_ORDER = ['id', 'check_definition_id', 'type', 'name', 'team', 'owning_team', 'responsible_team', 'description',
               'condition',
               'command', 'interval', 'entities', 'entities_exclude', 'status', 'last_modified_by']
FIELD_SORT_INDEX = {k: chr(i) for i, k in enumerate(FIELD_ORDER)}

LAST_MODIFIED_FMT = '%Y-%m-%d %H:%M:%S.%f'
LAST_MODIFIED_FMT_ZERO = '%Y-%m-%d %H:%M:%S'


class literal_unicode(str):
    '''Empty class to serialize value as literal YAML block'''
    pass


class CustomDumper(yaml.Dumper):
    '''Custom dumper to sort mapping fields as we like'''

    def represent_mapping(self, tag, mapping, flow_style=None):
        node = yaml.Dumper.represent_mapping(self, tag, mapping, flow_style)
        node.value = sorted(node.value, key=lambda x: FIELD_SORT_INDEX.get(x[0].value, x[0].value))
        return node


def literal_unicode_representer(dumper, data):
    node = dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return node


def remove_trailing_whitespace(text: str):
    '''Remove all trailing whitespace from all lines'''
    return '\n'.join([line.rstrip() for line in text.strip().split('\n')])


def dump_yaml(data):
    if isinstance(data, dict):
        for key, val in data.items():
            if key in LITERAL_FIELDS:
                # trailing whitespace would force YAML emitter to use doublequoted string
                data[key] = literal_unicode(remove_trailing_whitespace(val))

    return yaml.dump(data, default_flow_style=False, allow_unicode=True, Dumper=CustomDumper)


yaml.add_representer(literal_unicode, literal_unicode_representer)


def log_http_exception(e, act=None):
    err = act.error if act else error
    try:
        err('HTTP error: {} - {}'.format(e.response.status_code, e.response.reason))
        try:
            err(json.dumps(e.response.json(), indent=4))
        except Exception:
            err(e.response.text)
    except Exception:
        err('HTTP ERROR: {}'.format(e))


########################################################################################################################
# RENDERERS
########################################################################################################################
class Output:

    def __init__(self, msg, ok_msg=' OK', nl=False, output='text', pretty_json=False, printer=None,
                 suppress_exception=False):
        self.msg = msg
        self.ok_msg = ok_msg
        self.output = output
        self.nl = nl
        self.errors = []
        self.printer = printer
        self.indent = 4 if pretty_json else None
        self._suppress_exception = suppress_exception

    def __enter__(self):
        if self.output == 'text' and not self.printer:
            action(self.msg)
            if self.nl:
                secho('')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            if self.output == 'text' and not self.printer and not self.errors:
                ok(self.ok_msg)
        elif not self._suppress_exception:
            error(' EXCEPTION OCCURRED: {}'.format(exc_val))

    def error(self, msg, **kwargs):
        error(' {}'.format(msg), **kwargs)
        self.errors.append(msg)

    def echo(self, out):
        if self.output == 'yaml':
            print(dump_yaml(out))
        elif self.output == 'json':
            print(json.dumps(out, indent=self.indent))
        elif self.printer:
            self.printer(out, self.output)
        else:
            print(out)


def render_entities(entities, output):
    rows = []
    for e in entities:
        row = e
        s = sorted(e.keys())

        key_values = []

        for k in s:
            if k not in ('id', 'type'):
                if k == 'last_modified':
                    mod = row.pop('last_modified')
                    try:
                        row['last_modified_time'] = (
                            calendar.timegm(time.strptime(mod, LAST_MODIFIED_FMT)))
                    except ValueError:
                        row['last_modified_time'] = (
                            calendar.timegm(time.strptime(mod, LAST_MODIFIED_FMT_ZERO)))
                else:
                    key_values.append('{}={}'.format(k, e[k]))

        row['data'] = ' '.join(key_values)
        rows.append(row)

    rows.sort(key=lambda r: (r['last_modified_time'], r['id'], r['type']))

    with OutputFormat(output):
        print_table('id type last_modified_time data'.split(),
                    rows, titles={'last_modified_time': 'Modified'})


def render_status(status, output=None):
    secho('Alerts active: {}'.format(status.get('alerts_active')))

    info('Service Level Status:')
    in_tier = status.get('service_level_status', {}).get('ingest_max_check_tier', 3)
    if in_tier == 3:
        info("Ingest unlimited")
    elif in_tier == 2:
        info("Ingest only critial and important checks")
    else:
        info("Ingest only critial checks")

    q_tier = status.get('service_level_status', {}).get('query_max_check_tier', 3)
    q_dist = status.get('service_level_status', {}).get('query_distance_hours_limit', 0)
    if q_dist == 0:
        info("Query distance unlimited")
    else:
        info("Query distance limited to {} hours".format(q_dist))

    if q_tier == 3:
        info("Query metrics unlimited")
    elif q_tier == 2:
        info("Query metrics limited to critical and important checks")
    else:
        info("Query metrics limited to critical checks")


def render_checks(checks, output=None):
    rows = []

    for check in checks:
        row = check

        row['last_modified_time'] = calendar.timegm(time.gmtime(row.pop('last_modified') / 1000))

        row['name'] = row['name'][:60]
        row['owning_team'] = row['owning_team'][:60].replace('\n', '')

        rows.append(row)

    rows.sort(key=lambda c: c['id'])

    # Not really used since all checks are ACTIVE!
    check_styles = {
        'ACTIVE': {'fg': 'green'},
        'DELETED': {'fg': 'red'},
        'INACTIVE': {'fg': 'yellow'},
    }

    print_table(['id', 'name', 'owning_team', 'last_modified_time', 'last_modified_by', 'status', 'link'], rows,
                titles={'last_modified_time': 'Modified', 'last_modified_by': 'Modified by'}, styles=check_styles)


def render_alerts(alerts, output=None):
    rows = []

    for alert in alerts:
        row = alert

        row['last_modified_time'] = calendar.timegm(time.gmtime(row.pop('last_modified') / 1000))

        row['name'] = row['name'][:60]
        row['responsible_team'] = row['responsible_team'][:40].replace('\n', '')
        row['team'] = row['team'][:40].replace('\n', '')

        priorities = {1: 'HIGH', 2: 'MEDIUM', 3: 'LOW'}
        row['priority'] = priorities.get(row['priority'], 'LOW')

        rows.append(row)

    rows.sort(key=lambda c: c['id'])

    check_styles = {
        'ACTIVE': {'fg': 'green'},
        'REJECTED': {'fg': 'red'},
        'INACTIVE': {'fg': 'yellow'},
        'HIGH': {'fg': 'red'},
        'MEDIUM': {'fg': 'yellow', 'bold': True},
        'LOW': {'fg': 'yellow'},
    }

    titles = {
        'last_modified_time': 'Modified',
        'last_modified_by': 'Modified by',
        'check_definition_id': 'Check ID',
    }

    headers = [
        'id', 'name', 'check_definition_id', 'responsible_team', 'team', 'priority', 'last_modified_time',
        'last_modified_by', 'status', 'link',
    ]

    print_table(headers, rows, titles=titles, styles=check_styles)


def render_search(search, output):

    def _print_table(title, rows):
        info(title)
        rows.sort(key=lambda x: x.get('title'))
        print_table(['id', 'title', 'team', 'link'], rows)
        secho('')

    _print_table('Checks:', search['checks'])
    _print_table('Alerts:', search['alerts'])
    _print_table('Dashboards:', search['dashboards'])
    _print_table('Grafana Dashboards:', search['grafana_dashboards'])



import yaml

import click

from clickclick import AliasedGroup, Action, ok

from zmon_cli.cmds.command import (
    cli,
    get_client,
    yaml_output_option,
    pretty_json,
    output_option,
    retries,
)
from zmon_cli.output import dump_yaml, Output, render_checks
from zmon_cli.client import ZmonArgumentError


@cli.group('check-definitions', cls=AliasedGroup)
@click.pass_obj
def check_definitions(obj):
    """Manage check definitions"""
    pass


@check_definitions.command('init')
@click.argument('yaml_file', type=click.File('wb'))
def init(yaml_file):
    """Initialize a new check definition YAML file"""
    # NOTE: sorted like FIELD_SORT_INDEX
    name = click.prompt('Check definition name', default='Example Check')
    owning_team = click.prompt('Team owning this check definition (i.e. your team)', default='Example Team')

    data = {
        'name': name,
        'owning_team': owning_team,
        'description': "Example ZMON check definition which returns a HTTP status code.\n" +
                       "You can write multiple lines here, including unicode ☺",
        'command': "# GET request on example.org and return HTTP status code\n" +
                   "http('http://example.org/', timeout=5).code()",
        'interval': 60,
        'entities': [{'type': 'GLOBAL'}],
        'status': 'ACTIVE'
    }

    yaml_file.write(dump_yaml(data).encode('utf-8'))
    ok()


@check_definitions.command('get')
@click.argument('check_id', type=int)
@retries
@click.pass_obj
@yaml_output_option
@pretty_json
def get_check_definition(obj, check_id, retries, output, pretty):
    """Get a single check definition"""
    client = get_client(obj.config, retry_count=retries)

    with Output('Retrieving check definition ...', nl=True, output=output, pretty_json=pretty) as act:
        check = client.get_check_definition(check_id)

        keys = list(check.keys())
        for k in keys:
            if check[k] is None:
                del check[k]

        act.echo(check)


@check_definitions.command('list')
@click.pass_obj
@output_option
@pretty_json
def list_check_definitions(obj, output, pretty):
    """List all active check definitions"""
    client = get_client(obj.config)

    with Output('Retrieving active check definitions ...', nl=True, output=output, pretty_json=pretty,
                printer=render_checks) as act:
        checks = client.get_check_definitions()

        for check in checks:
            check['link'] = client.check_definition_url(check)

        act.echo(checks)


@check_definitions.command('filter')
@click.argument('field')
@click.argument('value')
@click.pass_obj
@output_option
@pretty_json
def filter_check_definitions(obj, field, value, output, pretty):
    """Filter active check definitions"""
    client = get_client(obj.config)

    with Output('Retrieving and filtering check definitions ...', nl=True, output=output, pretty_json=pretty,
                printer=render_checks) as act:
        checks = client.get_check_definitions()

        filtered = [check for check in checks if check.get(field) == value]

        for check in filtered:
            check['link'] = client.check_definition_url(check)

        act.echo(filtered)


@check_definitions.command('update')
@click.argument('yaml_file', type=click.File('rb'))
@click.option('--skip-validation', is_flag=True, help='Skip check command syntax validation.')
@retries
@click.pass_obj
def update(obj, yaml_file, skip_validation, retries):
    """Update a single check definition"""
    check = yaml.safe_load(yaml_file)

    check['last_modified_by'] = obj.get('user', 'unknown')

    client = get_client(obj.config, retry_count=retries)

    with Action('Updating check definition ...', nl=True) as act:
        try:
            check = client.update_check_definition(check, skip_validation=skip_validation)
            ok(client.check_definition_url(check))
        except ZmonArgumentError as e:
            act.error(str(e))


@check_definitions.command('delete')
@click.argument('check_id', type=int)
@click.pass_obj
def delete_check_definition(obj, check_id):
    """Delete an orphan check definition"""
    client = get_client(obj.config)

    with Action('Deleting check {} ...'.format(check_id)) as act:
        resp = client.delete_check_definition(check_id)

        if not resp.ok:
            act.error(resp.text)


@check_definitions.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



import os
import sys
import json
import yaml

import requests
import click

from clickclick import AliasedGroup, Action, action, ok, fatal_error

from zmon_cli.cmds.command import cli, get_client, output_option, yaml_output_option, pretty_json
from zmon_cli.output import render_entities, Output, log_http_exception

from zmon_cli.client import ZmonArgumentError

from calendar import timegm
from time import strptime


LAST_MODIFIED_FMT = '%Y-%m-%d %H:%M:%S.%f'
LAST_MODIFIED_FMT_ZERO = '%Y-%m-%d %H:%M:%S'


def entity_last_modified(e):
    try:
        return timegm(strptime(e.get('last_modified'), LAST_MODIFIED_FMT))
    except ValueError:
        return timegm(strptime(e.get('last_modified'), LAST_MODIFIED_FMT_ZERO))
    except Exception:
        return 0


########################################################################################################################
# ENTITIES
########################################################################################################################

@cli.group('entities', cls=AliasedGroup, invoke_without_command=True)
@click.pass_context
@output_option
@pretty_json
def entities(ctx, output, pretty):
    """Manage entities"""
    if not ctx.invoked_subcommand:
        client = get_client(ctx.obj.config)
        print("Fetching ALL entities, this might take some time\n", file=sys.stderr)

        with Output('Retrieving all entities ...', output=output, printer=render_entities, pretty_json=pretty) as act:
            entities = client.get_entities()
            act.echo(entities)


@entities.command('get')
@click.argument('entity_id')
@click.pass_obj
@yaml_output_option
@pretty_json
def get_entity(obj, entity_id, output, pretty):
    """Get a single entity by ID"""
    client = get_client(obj.config)

    with Output('Retrieving entity {} ...'.format(entity_id), nl=True, output=output, pretty_json=pretty) as act:
        entity = client.get_entity(entity_id)
        act.echo(entity)


@entities.command('filter')
@click.argument('filters', nargs=-1)
@click.pass_obj
@output_option
@pretty_json
def filter_entities(obj, filters, output, pretty):
    """
    List entities filtered by key values pairs

    E.g.:
        zmon entities filter type instance application_id my-app
    """
    client = get_client(obj.config)

    if len(filters) % 2:
        fatal_error('Invalid filters count: expected even number of args!')
    if len(filters) == 0:
        print("Fetching ALL entities, this might take some time\n", file=sys.stderr)

    with Output('Retrieving and filtering entities ...', nl=True, output=output, printer=render_entities,
                pretty_json=pretty) as act:

        query = dict(zip(filters[0::2], filters[1::2]))

        entities = client.get_entities(query=query)
        entities = sorted(entities, key=entity_last_modified)

        act.echo(entities)


@entities.command('push')
@click.argument('entity')
@click.pass_obj
def push_entity(obj, entity):
    """Push one or more entities"""
    client = get_client(obj.config)

    if (entity.endswith('.json') or entity.endswith('.yaml')) and os.path.exists(entity):
        with open(entity, 'rb') as fd:
            data = yaml.safe_load(fd)
    else:
        data = json.loads(entity)

    if not isinstance(data, list):
        data = [data]

    with Action('Creating new entities ...', nl=True) as act:
        for e in data:
            action('Creating entity {} ...'.format(e['id']))
            try:
                client.add_entity(e)
                ok()
            except ZmonArgumentError as e:
                act.error(str(e))
            except requests.HTTPError as e:
                log_http_exception(e, act)
            except Exception as e:
                act.error('Failed: {}'.format(str(e)))


@entities.command('delete')
@click.argument('entity_id')
@click.pass_obj
def delete_entity(obj, entity_id):
    """Delete a single entity by ID"""
    client = get_client(obj.config)

    with Action('Deleting entity {} ...'.format(entity_id)) as act:
        deleted = client.delete_entity(entity_id)
        if not deleted:
            act.error('Failed')


@entities.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())



from zmon_cli.cmds.command import cli

from zmon_cli.cmds.alert import alert_definitions
from zmon_cli.cmds.check import check_definitions
from zmon_cli.cmds.dashboard import dashboard
from zmon_cli.cmds.data import data
from zmon_cli.cmds.downtime import downtimes
from zmon_cli.cmds.entity import entities
from zmon_cli.cmds.grafana import grafana
from zmon_cli.cmds.group import groups, members
from zmon_cli.cmds.search import search
from zmon_cli.cmds.token import tv_tokens


__all__ = (
    alert_definitions,
    check_definitions,
    cli,
    dashboard,
    data,
    downtimes,
    entities,
    grafana,
    groups,
    members,
    search,
    tv_tokens,
)



import json

import yaml

import click

from clickclick import AliasedGroup, Action, ok

from zmon_cli.cmds.command import cli, get_client, yaml_output_option, output_option, pretty_json
from zmon_cli.output import dump_yaml, Output, render_alerts
from zmon_cli.client import ZmonArgumentError


@cli.group('alert-definitions', cls=AliasedGroup)
@click.pass_obj
def alert_definitions(obj):
    """Manage alert definitions"""
    pass


@alert_definitions.command('init')
@click.argument('yaml_file', type=click.File('wb'))
def init(yaml_file):
    """Initialize a new alert definition YAML file"""
    name = click.prompt('Alert name', default='Example Alert')
    check_id = click.prompt('Check ID')
    team = click.prompt('(Responsible-) Team', default='Example Team')

    data = {
        'check_definition_id': check_id,
        'condition': '>100',
        'description': 'Example Alert Description',
        'entities': [],
        'entities_exclude': [],
        'id': '',
        'name': name,
        'parameters': {},
        'parent_id': '',
        'priority': 2,
        'responsible_team': team,
        'status': 'ACTIVE',
        'tags': [],
        'team': team,
        'template': False,
    }

    yaml_file.write(dump_yaml(data).encode('utf-8'))
    ok()


@alert_definitions.command('get')
@click.argument('alert_id', type=int)
@click.pass_obj
@yaml_output_option
@pretty_json
def get_alert_definition(obj, alert_id, output, pretty):
    """Get a single alert definition"""
    client = get_client(obj.config)

    with Output('Retrieving alert definition ...', nl=True, output=output, pretty_json=pretty) as act:
        alert = client.get_alert_definition(alert_id)

        keys = list(alert.keys())
        for k in keys:
            if alert[k] is None:
                del alert[k]

        act.echo(alert)


@alert_definitions.command('list')
@click.pass_obj
@output_option
@pretty_json
def list_alert_definitions(obj, output, pretty):
    """List all active alert definitions"""
    client = get_client(obj.config)

    with Output('Retrieving active alert definitions ...', nl=True, output=output, pretty_json=pretty,
                printer=render_alerts) as act:
        alerts = client.get_alert_definitions()

        for alert in alerts:
            alert['link'] = client.alert_details_url(alert)

        act.echo(alerts)


@alert_definitions.command('filter')
@click.argument('field')
@click.argument('value')
@click.pass_obj
@output_option
@pretty_json
def filter_alert_definitions(obj, field, value, output, pretty):
    """Filter active alert definitions"""
    client = get_client(obj.config)

    with Output('Retrieving and filtering alert definitions ...', nl=True, output=output, pretty_json=pretty,
                printer=render_alerts) as act:
        alerts = client.get_alert_definitions()

        if field == 'check_definition_id':
            value = int(value)

        filtered = [alert for alert in alerts if alert.get(field) == value]

        for alert in filtered:
            alert['link'] = client.alert_details_url(alert)

        act.echo(filtered)


@alert_definitions.command('create')
@click.argument('yaml_file', type=click.File('rb'))
@click.pass_obj
def create_alert_definition(obj, yaml_file):
    """Create a single alert definition"""
    client = get_client(obj.config)

    alert = yaml.safe_load(yaml_file)

    alert['last_modified_by'] = obj.config.get('user', 'unknown')

    with Action('Creating alert definition ...', nl=True) as act:
        try:
            new_alert = client.create_alert_definition(alert)
            ok(client.alert_details_url(new_alert))
        except ZmonArgumentError as e:
            act.error(str(e))


@alert_definitions.command('update')
@click.argument('yaml_file', type=click.File('rb'))
@click.pass_obj
def update_alert_definition(obj, yaml_file):
    """Update a single alert definition"""
    alert = yaml.safe_load(yaml_file)

    alert['last_modified_by'] = obj.config.get('user', 'unknown')

    client = get_client(obj.config)

    with Action('Updating alert definition ...', nl=True) as act:
        try:
            # Workaround API inconsistency!
            if alert.get('parameters'):
                for k, v in alert['parameters'].items():
                    if type(v) is str:
                        alert['parameters'][k] = json.loads(v)

            client.update_alert_definition(alert)
            ok(client.alert_details_url(alert))
        except ZmonArgumentError as e:
            act.error(str(e))


@alert_definitions.command('delete')
@click.argument('alert_id', type=int)
@click.pass_obj
def delete_alert_definition(obj, alert_id):
    """Delete a single alert definition"""
    client = get_client(obj.config)

    with Action('Deleting alert definition ...'):
        client.delete_alert_definition(alert_id)


@alert_definitions.command('help')
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())
